#!/usr/bin/env perl
#
# This script is a part of Genome Search Toolkit
# Copyright (c) 2014 Kirill Kryukov
# see LICENSE.txt for details.
#

use strict;
use warnings;
use Getopt::Long qw(GetOptionsFromArray :config pass_through);
use FindBin;
use lib "$FindBin::Bin";
use gst_utils;

gst_init();

my ($format,$max_queries);
my ($min_score) = (0);
GetOptionsFromArray(\@opts,
    'min-score=f'   => \$min_score,
    'format=s'      => \$format,
    'max-queries=i' => \$max_queries
);
if (@opts) { die "Unrecognized option \"$opts[0]\"\n"; }

$help_str .= qq{Options:
    --format <FORMAT> - Output format string
    --min-score <S>   - Use only hits with at least score S
    --max-queries <N> - Process at most <N> queries
    --help            - Show help
    --version         - Show version
};
if (!defined($format) and !defined($summary_dir)) { $ver = 1; $help = 1; }
print_help();

if (!defined($format)) { die "Output format is not specified (--format option)\n"; }
if (!defined($summary_dir)) { die "Genome summary directory is not specified (--genome-summary-dir option)\n"; }

load_genome_taxonomy();

my @ids = ();
my %n_queries = ();
my ($qi,$qname,$score) = (0,'',0);

binmode STDIN;
binmode STDOUT;
while (<STDIN>)
{
    s/[\x0D\x0A]+$//;
    if (substr($_,0,1) eq ' ') { process(); if (defined($max_queries) and $qi >= $max_queries) { last; } $qname = substr($_,1); $qi++; next; }
    if ($qi<0) { next; }
    /^(\d+|\d+\.\d+)\s+(\S.+)$/ or die "Can't parse \"$_\"\n";
    my ($sc,$sname,$id) = ($1,$2,taxonomy_name_to_id($2));
    if ($sc < $min_score) { next; }
    if (!$id) { die "Unknown database name \"$sname\"\n"; }
    if ($sc > $score) { @ids = ($id); $score = $sc; }
    elsif (!scalar(@ids) or $sc == $score) { push @ids, $id; }
}
process();

$nodes{0}->{'n'} = '-';
print STDERR "# Number of taxon-pecific queries by taxon\n";
print STDERR "# non-cumulative\n";
print STDERR "# Columns: number of queries, taxonomic id, taxonomic name\n";
foreach my $id (sort { $n_queries{$b} <=> $n_queries{$a} || $nodes{$a}->{'n'} cmp $nodes{$b}->{'n'} } keys %n_queries)
{
    print STDERR $n_queries{$id}, "\t", $id, "\t", $nodes{$id}->{'n'}, "\n";
}

sub process
{
    if ($qname eq '') { return; }
    my $a = 0;
    if (scalar @ids)
    {
        $a = $ids[0];
        for (my $i=1; $i<scalar @ids; $i++) { $a = find_most_recent_common_ancestor($a,$ids[$i]); }
    }
    print format_output($a), "\n";
    $n_queries{$a}++;
    ($qname,$score) = ('',0);
    @ids = ();
}

sub format_output
{
    my ($sid) = @_;
    my $sname = ($sid > 0) ? $nodes{$sid}->{'n'} : '-';
    my $str = $format;
    $str =~ s/\{SID\}/$sid/g;
    $str =~ s/\{SNAME\}/$sname/g;
    $str =~ s/\{QINDEX\}/$qi/g;
    $str =~ s/\{QNAME\}/$qname/g;
    return $str;
}
