#!/usr/bin/env perl
#
# This script is a part of Genome Search Toolkit
# Copyright (c) 2014 Kirill Kryukov
# see LICENSE.txt for details.
#
# Cleans up and restructures a directory where each file corresponds to particular genome.
#     With '--delete' deletes files for missing genomes.
#     With '--move' moves files to match new structure.
#
# Collisions while moving files are resolved by deleting the source file and keeping the existing destination file.
# (Although deleting will only happen if '--delete' option is specified).
#

use strict;
use warnings;
use Cwd qw(abs_path);
use File::Basename qw(basename dirname);
use File::Copy;
use File::Find;
use File::Glob qw(:glob);
use File::Path qw(make_path remove_tree);
use File::Slurp;
use Getopt::Long qw(GetOptionsFromArray :config pass_through);
use Storable qw(retrieve);
use FindBin;
use lib "$FindBin::Bin";
use gst_utils;

my $default_struct_file = dirname(abs_path($0)) . '/struct-default.txt';

gst_init();

my ($dir_mask_list,$struct_file,$move,$delete,$dry,$verbose,$dirs,$extension_list);
my ($ok_to_delete_all,$ok_to_move_all,$ok_to_delete_or_move_all,$escape_spaces) = (0,0,0,0);
GetOptionsFromArray(\@opts,
    'dir=s'                    => \$dir_mask_list,
    'struct=s'                 => \$struct_file,
    'ext=s'                    => \$extension_list,
    'files'                    => sub { $dirs = 0; },
    'dirs'                     => sub { $dirs = 1; },
    'move'                     => \$move,
    'delete'                   => \$delete,
    'ok-to-delete-all'         => \$ok_to_delete_all,
    'ok-to-move-all'           => \$ok_to_move_all,
    'ok-to-delete-or-move-all' => \$ok_to_delete_or_move_all,
    'escape-spaces'            => \$escape_spaces,
    'dry'                      => \$dry,
    'verbose'                  => \$verbose
);
if (@opts) { die "Unrecognized option \"$opts[0]\"\n"; }

$help_str .= qq{Options:
    --dir <DIR>                - Directories to clean (one or several ;;-separated wildcards)
    --struct <S>               - Use directory structure defined in file <S>
    --ext <LIST>               - Comma-separated list of extensions to keep (only with --files)
    --use-db-taxa <LIST>       - Genome taxa to include (;;-separated)
    --files                    - Each genome is represented by a file
    --dirs                     - Each genome is represented by a directory
    --move                     - Can move files/directories
    --delete                   - Can delete files/directories
    --ok-to-delete-all         - OK to delete all files/directories
    --ok-to-move-all           - OK to move all files/directories
    --ok-to-delete-or-move-all - OK to delete of move all files/directories
    --escape-spaces            - Escape spaces in path
    --dry                      - Test without moving/deleting anything
    --help                     - Show help
    --version                  - Show version
};
if (!defined($dir_mask_list) or $dir_mask_list eq '') { $ver = 1; $help = 1; }
print_help();
if (!defined $dirs) { die "Either --files or --dirs must be specified\n"; }

if (!defined($struct_file)) { $struct_file = $default_struct_file; }
if (!-e $struct_file) { die "Can't find struct file \"$struct_file\"\n"; }

my @dir_list = ();
my @dir_masks = split(';;',$dir_mask_list);
foreach my $dir_mask (@dir_masks)
{
    $dir_mask =~ s/[\/\\]+$//;
    $dir_mask =~ s/\\/\//g;
    my $dir_mask_fixed = $dir_mask;
    $dir_mask_fixed =~ s/([\[\]])/\\$1/g;
    push @dir_list, bsd_glob($dir_mask_fixed);
}
@dir_list = grep { -d $_ } @dir_list;
if (!scalar(@dir_list)) { die "Can't find any directory matching \"$dir_mask_list\"\n"; }
my @extensions_to_keep = ();
if (defined $extension_list) { @extensions_to_keep = split(',',$extension_list); }

read_genome_info();
load_genome_taxonomy();
mark_genomes_to_use_by_taxa();

print STDERR "Reading taxonomy ..";
my $taxdb_nodes_hash_file = "$tax_db_dir/nodes.hash";
my %all_nodes = %{retrieve($taxdb_nodes_hash_file)};
print STDERR " ", scalar(keys %all_nodes), " nodes";
print STDERR ", hashing ..";
my (%all_name_to_taxid,%all_safe_name_to_taxid);
foreach my $taxid (keys %all_nodes)
{
    my $name = all_taxonomy_taxid_to_name($taxid);
    my $safe_name = change_name_to_safe_name($name);
    if (exists($all_name_to_taxid{$name})) { print STDERR "Duplicate name \"", $name, "\"\n"; }
    if (exists($all_safe_name_to_taxid{$safe_name})) { print STDERR "Same safe name: \"$safe_name\" (\"$name\", \"", $all_nodes{$all_safe_name_to_taxid{$safe_name}}->{'n'}, "\")\n"; }
    $all_name_to_taxid{$name} = $taxid;
    $all_safe_name_to_taxid{$safe_name} = $taxid;
}
print STDERR " OK\n";

if (defined $use_db_taxa)
{
    $all_nodes{1}->{'use'} = 0;
    foreach my $name (split(/;;/,$use_db_taxa))
    {
        my $use = 1;
        my $s = substr($name,0,1);
        if ($s eq '+') { $name = substr($name,1); }
        elsif ($s eq '-') { $name = substr($name,1); $use = 0; }
        my $id = all_taxonomy_name_to_taxid($name);
        if (!$id) { die "Unknown name: $name\n"; }
        $all_nodes{$id}->{'use'} = 1;
    }
}
else { $all_nodes{1}->{'use'} = 1; }

my @struct_dirs;
load_struct_dirs($struct_file);

my (%entries_to_delete,%entries_to_move,%dirs_to_delete_if_empty);
my ($n_entries_checked,$n_entries_deleted,$n_entries_moved) = (0,0,0);
my ($n_entries_checked_total,$n_entries_deleted_total,$n_entries_moved_total) = (0,0,0);
my ($dir,$dir_length) = ('',0);

for (@dir_list) { clean_directory($_); }

if ($n_entries_checked_total != $n_entries_checked) { print STDERR 'Total: ', format_stats($n_entries_checked_total,$n_entries_deleted_total,$n_entries_moved_total), "\n"; }
my $elapsed_time = time - $start_time;
print STDERR "All done, $elapsed_time second" . (($elapsed_time==1)?'':'s') . " elapsed\n";

sub clean_directory
{
    my ($d) = @_;
    print STDERR "Cleaning \"$d\"\n";
    if ($d eq '' or $d eq '/') { print STDERR "Can't process directory \"$d\"\n"; return; }

    %entries_to_delete = ();
    %entries_to_move = ();
    %dirs_to_delete_if_empty = ();
    ($n_entries_checked,$n_entries_deleted,$n_entries_moved) = (0,0,0);
    ($dir,$dir_length) = ($d,length($d));

    find({'wanted'=>\&{sub{ if (($dirs and -d and $_ ne $d) or (!$dirs and -f)) { check_entry($_); } }},'no_chdir'=>1},$d);

    if (scalar(keys %entries_to_delete) >= $n_entries_checked and !$ok_to_delete_all and !$ok_to_delete_or_move_all)
    {
        print STDERR "All entries in \"$d\" would be deleted\n - please add '--ok-to-delete-all' if this is correct.\n";
    }
    elsif (scalar(keys %entries_to_move) >= $n_entries_checked and !$ok_to_move_all and !$ok_to_delete_or_move_all)
    {
        print STDERR "All entries in \"$d\" would be moved\n - please add '--ok-to-move-all' if this is correct.\n";
    }
    elsif (scalar(keys %entries_to_delete) + scalar(keys %entries_to_move) >= $n_entries_checked and !$ok_to_delete_or_move_all)
    {
        print STDERR "All entries in \"$d\" would be moved or deleted\n - please add '--ok-to-delete-or-move-all' if this is correct.\n";
    }
    else
    {
        foreach my $e (sort keys %entries_to_delete)
        {
            if ($verbose) { print STDERR "Deleting $e\n"; }
            if (!$dry)
            {
                my $deleted_ok = 0;
                if ($dirs)
                {
                    if (!-d $e) { die "\"$e\" is not a directory\n"; }
                    my $err;
                    remove_tree($e,{'error' => \$err});
                    $deleted_ok = (scalar(@$err) == 0);
                }
                else { $deleted_ok = unlink $e; }
                if ($deleted_ok) { $dirs_to_delete_if_empty{dirname($e)} = 1; } else { die "Can't delete \"$e\"\n"; }
            }
            $n_entries_deleted++;
        }

        foreach my $from (sort keys %entries_to_move)
        {
            my $to = $entries_to_move{$from};
            if ($verbose) { print STDERR "Moving $from\n    => $to\n"; }
            if (!$dry)
            {
                my $to_dir = dirname($to);
                if (!-e $to_dir) { make_path($to_dir); }
                if (move($from,$to)) { $dirs_to_delete_if_empty{dirname($from)} = 1; } else { die "Can't move \"$from\" to \"$to\"\n"; }
            }         
            $n_entries_moved++;
        }

        for (keys %dirs_to_delete_if_empty) { delete_empty_dirs($_); }
    }

    print STDERR ' .. ', format_stats($n_entries_checked,$n_entries_deleted,$n_entries_moved), "\n";

    $n_entries_checked_total += $n_entries_checked;
    $n_entries_deleted_total += $n_entries_deleted;
    $n_entries_moved_total   += $n_entries_moved;
}

sub format_stats
{
    my ($n_checked,$n_deleted,$n_moved) = @_;
    my $s = 'Checked ' . commify($n_checked) . ' entries';
    if ($n_checked) { $s .= ': ' . commify($n_deleted) . ' deleted, ' . commify($n_moved) . ' moved'; }
    if ($dry and ($n_deleted or $n_moved)) { $s .= ' (dry run, actual files are not touched)'; }
    return $s;
}

sub check_entry
{
    my ($path) = @_;
    $path =~ s/\\/\//g;
    if (substr($path,0,$dir_length+1) ne "$dir/")
    {
        die "Path does not match the specified directory \"$path\"\n    " . substr($path,0,$dir_length+1) . "\n    $dir/\n";
    }

    my $base = basename($path);
    if (substr($base,0,1) eq '.') { return; }
    if ($dirs and $base =~ /^0-/) { return; }

    my $gname = $base;
    if (defined $extension_list)
    {
        my $ok = 0;
        for (@extensions_to_keep) { if ($base =~ /^(.+?)\.$_$/) { ($ok,$gname) = (1,$1); last; } }
        if (!$ok) { if ($delete) { $entries_to_delete{$path} = 1; } return; }
    }

    my $d = dirname($path);
    if ($escape_spaces) { $gname =~ s/_20/ /g; }

    if (exists $genomes{$gname})
    {
        if (!$genomes{$gname}->{'use'}) { return; }
        $n_entries_checked++;
        my $old_struct_dir = (length($d) > $dir_length+1) ? substr($d,$dir_length+1) : '';
        if (!exists $genomes{$gname}->{'diri'}) { die "Structure directory is not defined for \"$gname\"\n"; }
        my $correct_struct_dir = $struct_dirs[$genomes{$gname}->{'diri'}];
        if ($dirs) { $correct_struct_dir = tax_path_to_info_path($correct_struct_dir); }
        if ($escape_spaces) { $correct_struct_dir = escape_spaces($correct_struct_dir); }
        if ($old_struct_dir ne $correct_struct_dir)
        {
            my $to_path = "$dir/$correct_struct_dir/$base";
            if (-e $to_path)
            {
                if ($verbose and (-s $to_path != -s $path)) { print STDERR "Size mismatch:\n  (deleting) $path\n  (keeping)  $to_path\n"; }
                if ($delete) { $entries_to_delete{$path} = 1; }
            }
            else { if ($move) { $entries_to_move{$path} = "$dir/$correct_struct_dir/$base"; } }
        }
    }
    else
    {
        my $sp_name = $gname;
        $sp_name =~ s/\s*\[[^\[\]]+\]$//;
        $sp_name =~ s/\s*\[[^\[\]]+\]$//;
        my $taxid = all_taxonomy_name_to_taxid($sp_name);
        if ($taxid)
        {
            my $up = $taxid;
            while ($up != 1 and !exists($all_nodes{$up}->{'use'})) { $up = $all_nodes{$up}->{'p'}; }
            if (!exists($all_nodes{$up}->{'use'})) { die "Can't determine whether \"$sp_name\" should be used or not\n"; }
            if (!$all_nodes{$up}->{'use'}) { return; }
        }
        else { if ($verbose) { print STDERR "Unknown organism \"$sp_name\"\n"; } }
        $n_entries_checked++;

        if ($delete) { $entries_to_delete{$path} = 1; }
    }
}

sub load_struct_dirs
{
    my ($file) = @_;
    for (grep { !/^#/ and /=>/ } read_file(file_exists($file)))
    {
        /^\s*(\S.*?)\s*=>\s*(.*?)\s*$/ or die "Can't parse struct file line \"$_\"\n";
        my ($name,$d) = ($1,$2);
        my $id = taxonomy_name_to_id($name);
        if (!$id)
        {
            my $all_id = all_taxonomy_name_to_taxid($name);
            if (!$all_id) { die "Unknown name \"$name\" found in structure file \"$file\"\n"; }
        }
        $d =~ s/[\/\\]+$//;
        add_diri($id,scalar(@struct_dirs));
        push @struct_dirs, $d;
    }
}

sub add_diri
{
    my ($id,$n) = @_;
    for (@{$nodes{$id}->{'g'}}) { $genomes{$_}->{'diri'} = $n; }
    for (@{$nodes{$id}->{'c'}}) { add_diri($_,$n); }
}

sub all_taxonomy_taxid_to_name
{
    my ($taxid) = @_;
    if (!exists($all_nodes{$taxid})) { print STDERR "\nWarning: Trying to obtain a name for non-existing node \"$taxid\"\n"; return ''; }
    return $all_nodes{$taxid}->{'n'};
}

sub all_taxonomy_name_to_taxid
{
    my ($name) = @_;
    return exists($all_name_to_taxid{$name}) ? $all_name_to_taxid{$name} : exists($all_safe_name_to_taxid{$name}) ? $all_safe_name_to_taxid{$name} : 0;
}

sub delete_empty_dirs
{
    my ($d) = @_;
    if (!delete_empty_dir($d)) { delete_empty_dirs(dirname($d)); }
} 

sub delete_empty_dir
{
    my ($d) = @_;
    if ($d eq '' or $d eq '.' or $d =~ /[\/\\]$/) { return 1; }
    my @files = bsd_glob("$d/*");
    my $n = scalar @files;
    if (!$n) { rmdir($d); }
    return $n;
}
