#!/usr/bin/env perl
#
# This script is a part of Genome Search Toolkit
# Copyright (c) 2014 Kirill Kryukov
# see LICENSE.txt for details.
#
# Searches one query file on a cluster.
#

use strict;
use warnings;
use File::Basename qw(basename);
use File::Glob qw(:bsd_glob);
use File::Path qw(make_path);
use Getopt::Long qw(GetOptionsFromArray :config pass_through);
use FindBin;
use lib "$FindBin::Bin";
use gst_utils;

gst_init();

my ($clean,$qfile,$query_part_size,$genome_list_file,$n_genome_groups,$out_format,$opt_file,$qparts_dir,$gstk_dir,$merged_out_dir) = (0);
GetOptionsFromArray(\@opts,
    'query=s'            => \$qfile,
    'query-part-size=i'  => \$query_part_size,
    'genome-list-file=s' => \$genome_list_file,
    'n-genome-groups=i'  => \$n_genome_groups,
    'out-format=s'       => \$out_format,
    'opt-file=s'         => \$opt_file,
    'qparts-dir=s'       => \$qparts_dir,
    'gstk-dir=s'         => \$gstk_dir,
    'out-dir=s'          => \$merged_out_dir,
    'clean'              => \$clean
);
if (@opts) { die "Unrecognized option \"$opts[0]\"\n"; }

$help_str .= qq{Options:
    --query <PATH>            - Specify query file
    --query-part-size <N>     - Divide each query into parts of <N> bp
    --genome-list-file <FILE> - List of genomes to use
    --n-genome-groups <N>     - Divide genome database into <N> parts
    --out-format <TXT>        - Specify search output format
    --opt-file <FILE>         - Provide the options file
    --q-parts-dir <DIR>       - Directory for query parts ('gstk-qparts/<qname>' by default)
    --gstk-dir <DIR>          - Working directory ('gstk-<qname>-<opt>' by default)
    --out-dir <DIR>           - Directory with the eventual search output ('<gstk-dir>/out' by default)
    --use-db-taxa <LIST>      - Genome taxa to include (;;-separated)
    --clean                   - Clean search results before searching
    --help                    - Show help
    --version                 - Show version
};
if (!defined($qfile)) { $ver = 1; $help = 1; }
print_help();

if (!defined($out_format)) { die "Output format is not specified (--out-format option)\n"; }
if (!defined($opt_file)) { die "opt file is not specified (--opt-file option)\n"; }
if (!-e $opt_file or !-f $opt_file) { die "Can't find file \"$opt_file\"\n"; }
if (!defined($qparts_dir)) { $qparts_dir = "gstk-qparts"; }

if (!defined $n_genome_groups) { $n_genome_groups = 1; }
my $opt_name = defined($opt_file) ? $opt_file : '';
$opt_name =~ s/^opt-//;

my $qname = basename($qfile);
$qname =~ s/\.[^\.]+$//;

if (!defined($gstk_dir)) { $gstk_dir = "gstk-$qname-$opt_name"; }

my $query_option = "--query '$qfile'";
if (defined $query_part_size)
{
    my @part_files = bsd_glob("$qparts_dir/$qname/*");
    if (!-e "$qparts_dir/$qname" or !scalar(@part_files))
    {
        print_execute("fasta-splitter --out-dir $qparts_dir/$qname --part-size $query_part_size --measure seq --line-length 0 \"$qfile\"");
        @part_files = bsd_glob("$qparts_dir/$qname/*");
        if (!scalar(@part_files)) { die "Can't divide query file into parts\n"; }
    }
    $query_option = "--qdir '$qparts_dir/$qname'";
}

my ($sec,$min,$hour,$mday,$mon,$year,$wday,$yday,$isdst) = localtime();
my $run_id = sprintf("%04d-%02d-%02d-%02d%02d%02d",$year+1900,$mon+1,$mday,$hour,$min,$sec);

my $genome_lists_dir = "$gstk_dir/genome-lists/$run_id-$n_genome_groups";
my $scripts_dir      = "$gstk_dir/search-scripts/$run_id";
my $search_logs_dir  = "$gstk_dir/search-logs/$run_id";
my $qsub_logs_dir    = "$gstk_dir/qsub-logs/$run_id";
my $out_part_dir     = "$gstk_dir/out-by-part";
my $out_dir = defined($merged_out_dir) ? $merged_out_dir : "$gstk_dir/out";

if (-e $scripts_dir) { die "This run already exists\n"; }

if ($clean)
{
    print_execute("clean-genomes-directory --dir '$out_part_dir/*;;$out_dir' --ext $out_format --files --move --delete");
}

if (!-e $genome_lists_dir) { make_path($genome_lists_dir); }
unlink bsd_glob("$genome_lists_dir/*");

# Saving the list of genomes that already have merged output.
my $list_of_already_searched_genomes_file = "$genome_lists_dir/already-searched-genomes.txt";
print_execute("list-genomes-in-directory.pl --dir $out_dir --ext .$out_format >$list_of_already_searched_genomes_file");

# Saving the list of all genomes that should be searched (including those already searched earlier).
my $list_of_genomes_to_use_all_file = "$genome_lists_dir/genomes-to-use-all.txt";
my $opt_use_db_taxa = (defined($use_db_taxa) and $use_db_taxa ne '') ? "--use-db-taxa '$use_db_taxa'" : '';
if (defined $genome_list_file)
{
    if ($genome_list_file eq '') { die "Genome list file is not specified\n"; }
    if (!-e $genome_list_file) { die "Can't find genome list file\n"; }
    if (!-s $genome_list_file) { die "Genome list file is empty\n"; }
    print_execute("filter-genome-list-by-taxonomy.pl $opt_use_db_taxa <$genome_list_file >$list_of_genomes_to_use_all_file");
}
else
{
    print_execute("list-genomes.pl $opt_use_db_taxa >$list_of_genomes_to_use_all_file");
}

# Saving the list of genomes to use in this search (excluding those already searched).
my $list_of_genomes_to_use_this_time_file = "$genome_lists_dir/genomes-to-use-this-time.txt";
print_execute("filter-genome-list-keep-missing-in-another-list.pl" .
              " --list-file $list_of_already_searched_genomes_file" .
              " <$list_of_genomes_to_use_all_file" .
              " >$list_of_genomes_to_use_this_time_file");

# Dividing genome list into parts.
print_execute("partition-genome-list.pl --n-parts $n_genome_groups --prefix $genome_lists_dir/ --strict <$list_of_genomes_to_use_this_time_file");

make_path($scripts_dir);
make_path($search_logs_dir);
make_path($qsub_logs_dir);

unlink bsd_glob("$scripts_dir/*");
unlink bsd_glob("$search_logs_dir/*");
unlink bsd_glob("$qsub_logs_dir/*");

print STDERR "\nSubmitting $n_genome_groups jobs ";
for (my $g=1; $g<=$n_genome_groups; $g++)
{
    my $script_file = "$scripts_dir/g$g";
    open (my $S,'>',$script_file) or die "Can't create \"$script_file\"\n";
    binmode $S;
    print $S "for-each-query-and-genome --opt $opt_file $query_option --db-list-file $genome_lists_dir/$g" .
             " --db-order size-inc --outdir \"$out_part_dir\" --log $search_logs_dir/g$g.log";
    close $S;
    system("chmod a+x $script_file");
    `qsub -V -N gstk-g$g -d $start_dir -j oe -o $qsub_logs_dir/g$g.log -l nodes=1:ppn=12,walltime=720:00:00 $script_file`;
    print STDERR '.';
}
print STDERR " OK\n";

sub print_execute { my ($cmd) = @_; print STDERR "\n$cmd\n"; system($cmd); }
