#!/usr/bin/env perl
#
# This script is a part of Genome Search Toolkit
# Copyright (c) 2016 Kirill Kryukov
# see LICENSE.txt for details.
#

use strict;
use warnings;
use Getopt::Long qw(GetOptionsFromArray :config pass_through);
use FindBin;
use lib "$FindBin::Bin";
use gst_utils;

gst_init();

my ($max_queries);
my ($min_score) = (0);
GetOptionsFromArray(\@opts,
    'min-score=f'   => \$min_score,
    'max-queries=i' => \$max_queries
);
if (@opts) { die "Unrecognized option \"$opts[0]\"\n"; }

$help_str .= qq{Options:
    --min-score <S>   - Use only hits with at least score S
    --max-queries <N> - Process at most <N> queries
    --help            - Show help
    --version         - Show version
};
if (!defined($summary_dir)) { $ver = 1; $help = 1; }
print_help();

if (!defined($summary_dir)) { die "Genome summary directory is not specified (--genome-summary-dir option)\n"; }

load_genome_taxonomy();

# Constructing hash for looking up taxonomic ID by name.
my %taxid_to_name;
foreach my $n (keys %name_to_taxid) { $taxid_to_name{$name_to_taxid{$n}} = $n; }
foreach my $n (keys %safe_name_to_taxid) { $taxid_to_name{$safe_name_to_taxid{$n}} = $n; }
my %all_name_to_taxid;
foreach my $t (keys %taxid_to_name) { $all_name_to_taxid{$taxid_to_name{$t}} = $t; }

my $n_queries_with_no_hits = 0;
my ($n_hits,%n_hits_by_taxon,%n_queries) = (0);
my ($qi) = (0);

binmode STDIN;
binmode STDOUT;
while (<STDIN>)
{
    s/[\x0D\x0A]+$//;
    if (substr($_,0,1) eq ' ') { process(); if (defined($max_queries) and $qi >= $max_queries) { last; } $qi++; next; }
    my $i = index($_,' ');
    if ($i < 0) { die "Can't parse \"$_\"\n"; }
    my $sc = substr($_,0,$i);
    if ($sc < $min_score) { next; }
    my $sname = substr($_,$i+1);
    if (!exists $all_name_to_taxid{$sname}) { die "Unknown name: \"$sname\"\n"; }
    my $id = $all_name_to_taxid{$sname};
    $n_hits++;
    $n_hits_by_taxon{$id}++;
}
process();

sub process
{
    if ($qi < 1) { return; }
    if ($n_hits == 0) { $n_queries_with_no_hits++; return; }
    my $n = 1/$n_hits;
    foreach my $id (keys %n_hits_by_taxon) { $n_queries{$id} += $n_hits_by_taxon{$id} * $n; }
    %n_hits_by_taxon = ();
    $n_hits = 0;
}

print "# Number of queries by taxon\n";
print "# Columns: number of queries, taxonomic id, taxonomic name\n";
if ($n_queries_with_no_hits > 0) { print sprintf("%.6f",$n_queries_with_no_hits), "\t0\t-\n"; }
foreach my $id (sort { $n_queries{$b} <=> $n_queries{$a} || $nodes{$a}->{'n'} cmp $nodes{$b}->{'n'} } keys %n_queries)
{
    print sprintf("%.6f",$n_queries{$id}), "\t", $id, "\t", $nodes{$id}->{'n'}, "\n";
}

my $elapsed_time = time - $start_time;
print STDERR " - done, $elapsed_time second" . (($elapsed_time==1)?'':'s') . " elapsed\n";
