#!/usr/bin/env perl
#
# This script is a part of Genome Search Toolkit
# Copyright (c) 2016 Kirill Kryukov
# see LICENSE.txt for details.
#

#
# Input: Non-cumulative count-per-taxon file.
# Output: Cumulative count-per-taxon file.
#

use strict;
use warnings;
use Getopt::Long qw(GetOptionsFromArray :config pass_through);
use FindBin;
use lib "$FindBin::Bin";
use gst_utils;

gst_init();

my ($strict) = (0);
GetOptionsFromArray(\@opts,
    'strict' => \$strict
);
if (@opts) { die "Unrecognized option \"$opts[0]\"\n"; }

$help_str .= qq{Options:
    --strict   - Only work if all taxa are recognized
    --help     - Show help
    --version  - Show version
};
print_help();

load_genome_taxonomy();

my ($header,$cumulative_specified,$cumulative_str) = ('',0,"# cumulative\n");
my ($correct_n_fields,$n_datasets,$n_fractional) = (0,0,0);
my (%unknown_nodes,%n_queries_per_taxid,%n_queries_per_taxid_cumulative,%sum);

binmode(STDIN);
while (<STDIN>)
{
    s/[\x0D\x0A]+$//;
    if (substr($_,0,1) eq '#')
    {
        if (/#\s*cumulative/i) { if ($strict) { die "The input is already cumulative\n"; } }
        elsif (/#\s*non-cumulative/i) { if (!$cumulative_specified) { $header .= $cumulative_str; $cumulative_specified = 1; } }
        else { $header .= "$_\n"; }
        next;
    }

    my @fields = split("\t",$_);
    my $n_fields = scalar(@fields);
    if ($correct_n_fields == 0)
    {
        if ($n_fields < 3) { die "Input has too few columns ($n_fields), there should be at least 3\n"; }
        $correct_n_fields = $n_fields;
        $n_datasets = $n_fields - 2;
    }
    elsif ($n_fields != $correct_n_fields) { die "Can't find the expected number ($correct_n_fields) of fields in \"$_\"\n"; }

    my $taxid = $fields[$n_datasets];
    if ($taxid != 0 and !exists $nodes{$taxid})
    {
        my $taxname = $fields[$n_fields-1];
        if ($strict) { die "Unknown taxon: $taxid, \"$taxname\"\n"; }
        else { $unknown_nodes{$taxid} = $taxname; next; }
    }

    for (my $d = 0; $d < $n_datasets; $d++)
    {
        $n_queries_per_taxid{$taxid}->[$d] = $fields[$d];
        if ($fields[$d] =~ /\./) { $n_fractional++; }
    }
}

foreach my $leaf (keys %n_queries_per_taxid)
{
    if ($leaf == 0) { next; }
    my @data = @{$n_queries_per_taxid{$leaf}};
    for (my $t = $leaf; $t; $t = $nodes{$t}->{'p'})
    {
        for (my $d = 0; $d < $n_datasets; $d++) { $n_queries_per_taxid_cumulative{$t}->[$d] += $data[$d]; }
    }
}

if (scalar keys %unknown_nodes) { print STDERR scalar(keys %unknown_nodes), " unknown nodes found\n"; }

binmode(STDOUT);
if (!$cumulative_specified) { $header = $cumulative_str . $header; }
print $header;

foreach my $taxid (keys %n_queries_per_taxid_cumulative)
{
    for (my $d = 0; $d < $n_datasets; $d++) { $sum{$taxid} += $n_queries_per_taxid_cumulative{$taxid}->[$d]; }
}

my $format = $n_fractional ? '%.6f' : '%.0f';

if (exists $n_queries_per_taxid{0}) { print join("\t", map { sprintf($format,$_); } @{$n_queries_per_taxid{0}}), "\t0\t-\n"; }
foreach my $taxid (sort { $sum{$b} <=> $sum{$a} } keys %n_queries_per_taxid_cumulative)
{
    print join("\t", map { sprintf($format,$_); } @{$n_queries_per_taxid_cumulative{$taxid}}), "\t$taxid\t", $nodes{$taxid}->{'n'}, "\n";
}
