#!/usr/bin/env perl
#
# This script is a part of Genome Search Toolkit
# Copyright (c) 2016 Kirill Kryukov
# see LICENSE.txt for details.
#

#
# Input: count-per-taxon file with multiple datasets
# Output: count-per-taxon file with one dataset
#

use strict;
use warnings;
use Getopt::Long qw(GetOptionsFromArray :config pass_through);
use FindBin;
use lib "$FindBin::Bin";
use gst_utils;

gst_init();

my ($d_index,$d_name,$strict) = (0,'',0);
GetOptionsFromArray(\@opts,
    'dataset-index=i' => \$d_index,
    'dataset-name=s'  => \$d_name,
    'strict'          => \$strict
);
if (@opts) { die "Unrecognized option \"$opts[0]\"\n"; }

$help_str .= qq{Options:
    --dataset-index N - Extract dataset with index N (1-based)
    --dataset-name S  - Extract dataset named S
    --strict          - Only work if all taxa are recognized
    --help            - Show help
    --version         - Show version
};
if ($d_index > 0 and $d_name ne '') { die "Either --dataset-name or --dataset-index should be specified, but not both\n"; }
if ($d_index <= 0 and $d_name eq '') { $ver = 1; $help = 1; }
print_help();

my ($correct_n_fields,$col_index) = (0,-1);
my ($header,$col_names_str,$col_name) = ('','','');

binmode(STDIN);
binmode(STDOUT);
while (<STDIN>)
{
    s/[\x0D\x0A]+$//;
    if (substr($_,0,1) eq '#') { if ($col_names_str ne '') { $header .= "$col_names_str\n"; } $col_names_str = $_; next; }

    if ($col_index < 0)
    {
        $col_names_str =~ s/^#\s*//;
        my @col_names = split("\t",$col_names_str);
        $correct_n_fields = scalar(@col_names);
        if ($correct_n_fields < 3) { die "Input has too few columns ($correct_n_fields), there should be at least 3\n"; }
        my $n_datasets = $correct_n_fields - 2;

        if ($d_index <= 0)
        {
            for (my $d = 0; $d < $n_datasets; $d++)
            {
                if ($col_names[$d] eq $d_name) { $col_index = $d; $col_name = $d_name; last; }
            }
            if ($col_index < 0) { die "Can't find dataset \"$d_name\" in the input\n"; }
        }
        else
        {
            if ($d_index > $n_datasets) { die "Requested dataset with index $d_index, but input has only $n_datasets datasets\n"; }
            $col_index = $d_index - 1;
            $col_name = $col_names[$col_index];
        }

        print $header;
        print "# $col_name\t", $col_names[$n_datasets], "\t", $col_names[$n_datasets+1], "\n";
    }

    my @fields = split("\t",$_);
    my $n_fields = scalar(@fields);
    if ($n_fields != $correct_n_fields) { die "Can't find the expected number ($correct_n_fields) of fields in \"$_\"\n"; }

    my $n = $fields[$col_index];
    if ($n > 0)
    {
        print $n, "\t", $fields[$n_fields-2], "\t", $fields[$n_fields-1], "\n";
    }
}
