#!/usr/bin/env perl
#
# This script is a part of Genome Search Toolkit
# Copyright (c) 2016 Kirill Kryukov
# see LICENSE.txt for details.
#

#
# Input: Cumulative count-per-taxon file.
# Output: Non-cumulative count-per-taxon file.
#

use strict;
use warnings;
use Getopt::Long qw(GetOptionsFromArray :config pass_through);
use FindBin;
use lib "$FindBin::Bin";
use gst_utils;

gst_init();

my ($strict) = (0);
GetOptionsFromArray(\@opts,
    'strict' => \$strict
);
if (@opts) { die "Unrecognized option \"$opts[0]\"\n"; }

$help_str .= qq{Options:
    --strict   - Only work if all taxa are recognized
    --help     - Show help
    --version  - Show version
};
print_help();

load_genome_taxonomy();

my ($header,$cumulative_specified,$cumulative_str) = ('',0,"# non-cumulative\n");
my ($correct_n_fields,$n_datasets,$n_fractional) = (0,0,0);
my (%unknown_nodes,%n_queries_per_taxid,%sum);

binmode(STDIN);
while (<STDIN>)
{
    s/[\x0D\x0A]+$//;
    if (substr($_,0,1) eq '#')
    {
        if (/#\s*non-cumulative/i) { if ($strict) { die "The input is already non-cumulative\n"; } }
        elsif (/#\s*cumulative/i) { if (!$cumulative_specified) { $header .= $cumulative_str; $cumulative_specified = 1; } }
        else { $header .= "$_\n"; }
        next;
    }

    my @fields = split("\t",$_);
    my $n_fields = scalar(@fields);
    if ($correct_n_fields == 0)
    {
        if ($n_fields < 3) { die "Input has too few columns ($n_fields), there should be at least 3\n"; }
        $correct_n_fields = $n_fields;
        $n_datasets = $n_fields - 2;
    }
    elsif ($n_fields != $correct_n_fields) { die "Can't find the expected number ($correct_n_fields) of fields in \"$_\"\n"; }

    my $taxid = $fields[$n_datasets];
    if ($taxid != 0 and !exists $nodes{$taxid})
    {
        my $taxname = $fields[$n_fields-1];
        if ($strict) { die "Unknown taxon: $taxid, \"$taxname\"\n"; }
        else { $unknown_nodes{$taxid} = $taxname; next; }
    }

    for (my $d = 0; $d < $n_datasets; $d++)
    {
        $n_queries_per_taxid{$taxid}->[$d] = $fields[$d];
        if ($fields[$d] =~ /\./) { $n_fractional++; }
    }
}

uncumulate(1);

sub uncumulate
{
    my ($id) = @_;
    for (@{$nodes{$id}->{'c'}}) { if (exists $n_queries_per_taxid{$_}) { uncumulate($_); } }
    for (my $d = 0; $d < $n_datasets; $d++)
    {
        my $n = $n_queries_per_taxid{$id}->[$d];
        if ($n <= 0) { next; }
        for (my $t = $nodes{$id}->{'p'}; $t; $t = $nodes{$t}->{'p'})
        {
            $n_queries_per_taxid{$t}->[$d] -= $n;
            if ($n_queries_per_taxid{$t}->[$d] < 0) { $n_queries_per_taxid{$t}->[$d] = 0; }
        }
    }
}

if (scalar keys %unknown_nodes) { print STDERR scalar(keys %unknown_nodes), " unknown nodes found\n"; }

binmode(STDOUT);
if (!$cumulative_specified) { $header = $cumulative_str . $header; }
print $header;

foreach my $taxid (keys %n_queries_per_taxid)
{
    for (my $d = 0; $d < $n_datasets; $d++) { $sum{$taxid} += $n_queries_per_taxid{$taxid}->[$d]; }
}

my $format = $n_fractional ? '%.6f' : '%.0f';

if (exists $n_queries_per_taxid{0}) { print join("\t", map { sprintf($format,$_); } @{$n_queries_per_taxid{0}}), "\t0\t-\n"; }
foreach my $taxid (sort { $sum{$b} <=> $sum{$a} || $nodes{$a}->{'n'} cmp $nodes{$b}->{'n'} } grep { $sum{$_} > 0 } grep { $_ != 0 } keys %n_queries_per_taxid)
{
    print join("\t", map { sprintf($format,$_); } @{$n_queries_per_taxid{$taxid}}), "\t$taxid\t", $nodes{$taxid}->{'n'}, "\n";
}
