#!/usr/bin/env perl
#
# This script is a part of Genome Search Toolkit
# Copyright (c) 2015 Kirill Kryukov
# see LICENSE.txt for details.
#
# Extracts database genome regions that are homologous to query sequences, based on search output in b6 format.
#
# Saves homologs of each query sequence separately from each other.
# Does not merge overlapping homologs, saves them as separate seqeunces.
# Cares about the strand, saves sequence in the strand that aligned to query.
#
# Stores regions in fasta format, with names constructed as:
#    Name-of-query-sequence Genome-name Index-of-genome-sequence Strand Start-End Score
# (Start and end coordinates are 1-based and inclusive on both ends).
#

use strict;
use warnings;
use File::Glob ':bsd_glob';
use File::Path qw(make_path);
use File::Slurp;
use Getopt::Long qw(GetOptionsFromArray :config pass_through);
use FindBin;
use lib "$FindBin::Bin";
use gst_utils;

gst_init();

my ($min_score,$min_pident,$min_length,$verbose,$in_dir_mask,$namefmt,$specific_query_name) = (0,0,0,0);
GetOptionsFromArray(\@opts,
    'in=s'         => \$in_dir_mask,
    'min-score=f'  => \$min_score,
    'min-pident=f' => \$min_pident,
    'min-length=i' => \$min_length,
    'namefmt=s'    => \$namefmt,
    'query=s'      => \$specific_query_name,
    'verbose'      => \$verbose
);
if (@opts) { die "Unrecognized option \"$opts[0]\"\n"; }

$help_str .= qq{Options:
    --in <MASK>          - Specify a mask for input directories
    --min-score <S>      - Use only hits with at least this score
    --min-pident <I>     - Use only hits with at least this percent identity
    --min-length <L>     - Output only regions of at least this length
    --namefmt <FORMAT>   - Sequence name format, may include: \$qname,
                            \$sgname, \$staxname, \$sname, \$sorgname,
                            \$sstart, \$send, \$strand, \$score,
                            \$hitindexq
    --query <NAME>       - Only extract homologs of one query sequence
    --use-db-taxa <LIST> - Genome taxa to include (;;-separated)
    --help               - Show help
    --version            - Show version
};
print_help();
if (!defined($namefmt)) { $namefmt = '$qname $sgname $sname $strand $sstart-$send $score'; }
if (!defined($in_dir_mask)) { die "Input directories are not specified (--in option)\n"; }

my $in_dir_mask_fix = $in_dir_mask;
$in_dir_mask_fix =~ s/([\[\]])/\\$1/g;
my @in_dirs = bsd_glob($in_dir_mask_fix);
my $n_in_dirs = scalar @in_dirs;
if (!$n_in_dirs) { die "Can't find any directories matching \"$in_dir_mask\"\n"; }

read_genome_info();
if (defined($use_db_taxa)) { load_genome_taxonomy(); }
mark_genomes_to_use_by_taxa();

my $n_selected_genomes = scalar grep { $genomes{$_}->{'use'} } keys %genomes;
print STDERR "Extracting homologous regions from $n_selected_genomes genomes based on output from $n_in_dirs directories\n";

my ($n_genomes_processed,$n_files_found,$n_files_processed,$n_extracted) = (0,0,0,0);

foreach my $gname (sort { $genomes{$a}->{'tax_path'} cmp $genomes{$b}->{'tax_path'} || $a cmp $b } grep { $genomes{$_}->{'use'} } keys %genomes)
{
    my $tax_path = $genomes{$gname}->{'tax_path'};
    my $in_rel_path = "$tax_path/$gname.b6";

    my $n_inputs = 0;
    foreach my $in_dir (@in_dirs) { if (-e "$in_dir/$in_rel_path") { $n_inputs++; } }

    if (!$n_inputs) { next; }
    elsif ($n_inputs < $n_in_dirs) { next; }
    elsif ($n_inputs != $n_in_dirs) { die "Unexpected number of inputs: $n_inputs\n"; }

    $n_genomes_processed++;
    if ($verbose) { print STDERR "$tax_path/$gname\n"; }
    if ($verbose) { if ($n_inputs != 1) { print STDERR "    $n_inputs inputs -> merging\n"; } }

    my $taxname = $gname;
    $taxname =~ s/\s+\[.+?\]//g;
    my $info_path = tax_path_to_info_path($tax_path);
    my $blast_db_path = "$blast_db_dir/" . escape_spaces("$info_path/$gname") . "/db";
    my @hits;

    my %n_hits_by_query;

    foreach my $in_dir (@in_dirs)
    {
        $n_files_found++;
        my $in_path = "$in_dir/$in_rel_path";
        if (-s $in_path == 0) { next; }
        $n_files_processed++;

        #print STDERR "\n===== $in_path\n";

        my $info_path = tax_path_to_info_path($tax_path);
        my @original_names = read_file(file_exists("$info_dir/$info_path/$gname/names"), 'binmode' => ':raw', 'chomp' => 1);

        open(my $IN,'<',$in_path) or die "Can't open \"$in_path\"\n";
        binmode $IN;
        while (<$IN>)
        {
            s/[\x0D\x0A]+$//;
            my @fields = split(/\s+/,$_);
            if (scalar(@fields) != 12) { die "Can't parse \"$_\" in \"$in_path\"\n"; }

            my ($qid,$sid,$pident,$sstart,$send,$as) = @fields[0,1,2,8,9,11];
            if (defined $specific_query_name and $qid ne $specific_query_name) { next; }

            my ($bstart,$bend) = ($sstart <= $send) ? ($sstart,$send) : ($send,$sstart);
            my $slen = $bend - $bstart + 1;

            if ($as >= $min_score and $pident >= $min_pident and $slen >= $min_length)
            {
                my $strand = ($sstart <= $send) ? 'plus' : 'minus';
                my $strnd = ($sstart <= $send) ? '+' : '-';

                my $cmd = "blastdbcmd -db \"$blast_db_path\" -entry \"seq" . $sid . "\" -range $bstart-$bend -strand $strand -outfmt %s";
                my $seq;
                {
                    open(my $P,'-|',$cmd) or die $!;
                    binmode $P;
                    $seq = <$P>;
                    close $P;
                }
                $seq =~ s/\s+$//;

                if (!defined($seq) or $seq eq '') { die "Failed to extract sequence:\n$cmd\n"; }

                my $sorgname = ($sid <= scalar(@original_names)) ? substr($original_names[$sid-1],1) : '-';

                my $name = $namefmt;
                $name =~ s/\$sstart/$bstart/g;
                $name =~ s/\$send/$bend/g;
                $name =~ s/\$score/$as/g;
                $name =~ s/\$strand/$strnd/g;
                $name =~ s/\$qname/$qid/g;
                $name =~ s/\$staxname/$taxname/g;
                $name =~ s/\$sgname/$gname/g;
                $name =~ s/\$sname/$sid/g;
                $name =~ s/\$sorgname/$sorgname/g;

                push @hits, { 'name' => $name, 'seq' => $seq, 'score' => $as, 'qid' => $qid, 'staxname' => $taxname };
                $n_extracted++;
                $n_hits_by_query{$qid}++;
                if (!$verbose) { print STDERR ">"; }
            }
            else { if (!$verbose) { print STDERR "."; } }
        }
        close $IN;
    }

    my %hit_index_by_query;
    foreach my $hit (sort { $b->{'score'} <=> $a->{'score'} } @hits)
    {
        my ($qid,$name) = ($hit->{'qid'},$hit->{'name'});
        if ($n_hits_by_query{$qid} > 1)
        {
            my $i = exists($hit_index_by_query{$qid}) ? $hit_index_by_query{$qid}+1 : 1;
            $hit_index_by_query{$qid} = $i;
            $name =~ s/\$hitindexq/$i/g;
        }
        else { $name =~ s/\s+\$hitindexq//g; }
        print ">$name\n";
        #print $hit->{'seq'}, "\n";
        while ($hit->{'seq'} =~ /(.{1,100})/g) { print $1, "\n"; }
    }
}
if (!$verbose and $n_extracted > 0) { print STDERR "\n"; }

print STDERR "Found $n_files_found files, processed $n_files_processed non-empty files\n";
print STDERR "Extracted $n_extracted sequences\n";
my $elapsed_time = time - $start_time;
print STDERR "$elapsed_time second" . (($elapsed_time==1)?'':'s') . " elapsed\n";
