#
# This script is a part of Genome Search Toolkit
# Copyright (c) 2014 Kirill Kryukov
# see LICENSE.txt for details.
#

package gst_utils;
use strict;
use warnings;
use Cwd qw(cwd abs_path);
use Exporter qw(import);
use File::Basename qw(basename dirname);
use File::Glob ':bsd_glob';
use File::Path qw(make_path);
use File::Slurp;
use File::Util;
use FindBin;
use Getopt::Long qw(GetOptionsFromArray :config pass_through);
use List::MoreUtils qw(first_index);
use List::Util qw(sum);
use Parallel::ForkManager;
use Storable qw(retrieve);

our @EXPORT = qw(
    $sep $start_dir $start_time $script_name
    $ver_str $help_str

    @opts
    $scripts_dir $summary_dir $info_dir $bz2_dir $fasta_dir $blast_db_dir $blat_db_dir $bwa_db_dir $last_db_dir
    $tax_db_dir $cdata_dir $cgraphs_dir $cwords_dir
    $krona_tools_dir
    $ntasks $nthreads $ntasks_small $nthreads_small $small_db_fasta_size
    $use_db_taxa $help $ver

    %genomes
    %nodes %name_to_taxid %safe_name_to_taxid

    &gst_init &print_help
    &read_genome_info
    &report_about_loaded_genomes

    &load_genome_taxonomy
    &taxonomy_name_to_id
    &find_most_recent_common_ancestor
    &mark_genomes_to_use_by_taxa
    &set_genome_use_flag_in_subtree

    &tax_path_to_info_path

    &process_stdinout_or_dir
    &find_files

    &change_name_to_safe_name
    &unescape_file_name

    &escape_spaces
    &fix_separator &unix_separator
    &file_exists &dir_exists
    &file_mtime
    &commify
);

our ($sep,$start_dir,$start_time,$script_name) = (File::Util->SL,cwd(),time,$FindBin::Script);
our ($package_name,$package_version,$package_date,$package_years) = ('genome-search-toolkit','0.1.1','2016-03-03','2016');

our @opts;
our ($scripts_dir,$summary_dir,$info_dir,$bz2_dir,$fasta_dir,$blast_db_dir,$blat_db_dir,$bwa_db_dir,$last_db_dir) = ('','','','','','','','','');
our ($krona_tools_dir) = ('');
our ($tax_db_dir,$cdata_dir,$cgraphs_dir,$cwords_dir) = ('','','','');
our ($use_db_taxa,$help,$ver);
our ($ntasks,$nthreads,$ntasks_small,$nthreads_small,$small_db_fasta_size) = (1,1,1,1,0);

our $ver_str = "$package_name $package_version, $package_date\nCopyright (c) $package_years Kirill Kryukov\n";;
our $help_str = "Usage: $script_name <option>...\n";

our %genomes;
our (%nodes,%name_to_taxid,%safe_name_to_taxid);

sub gst_init
{
    $| = 1;
    if ($sep eq '\\') { $ENV{'CYGWIN'} = 'nodosfilewarning'; }

    @opts = @ARGV;
    my %opt_file_used;
    for ('.',dirname(abs_path($0)))
    {
        if (!-e "$_/genome-search-toolkit.conf") { next; }
        my $p = fix_separator(abs_path("$_/genome-search-toolkit.conf"));
        if (-e $p) { unshift @opts, ('--opt',$p); }
    }
    while ((my $i = (first_index { /^--opt$/i; } @opts[0..$#opts-1])) >= 0)
    {
        my $file = fix_separator(abs_path(file_exists($opts[$i+1])));
        if (exists $opt_file_used{$file}) { splice @opts, $i, 2; }
        else
        {
            #print "Loading \"$file\"\n";
            $opt_file_used{$file} = 1;
            splice @opts, $i, 2, map { /^\s*(\S+)\s+(\S.*?)\s*$/; ('--'.$1,$2); } grep {!/^#/ and /\S\s+\S/ } read_file($file);
        }
    }

    $scripts_dir = dirname($script_name);

    GetOptionsFromArray(\@opts,
        'genome-summary-dir=s'     => \$summary_dir,
        'genome-info-dir=s'        => \$info_dir,
        'genome-bz2-dir=s'         => \$bz2_dir,
        'genome-fasta-dir=s'       => \$fasta_dir,
        'genome-blast-db-dir=s'    => \$blast_db_dir,
        'genome-blat-db-dir=s'     => \$blat_db_dir,
        'genome-bwa-db-dir=s'      => \$bwa_db_dir,
        'genome-last-db-dir=s'     => \$last_db_dir,
        'tax-db-dir=s'             => \$tax_db_dir,
        'genome-comp-data-dir=s'   => \$cdata_dir,
        'genome-comp-graphs-dir=s' => \$cgraphs_dir,
        'genome-comp-words-dir=s'  => \$cwords_dir,
        'krona-tools-dir=s'        => \$krona_tools_dir,
        'ntasks=i'                 => \$ntasks,
        'nthreads=i'               => \$nthreads,
        'ntasks-small=i'           => \$ntasks_small,
        'nthreads-small=i'         => \$nthreads_small,
        'small-db-fasta-size=i'    => \$small_db_fasta_size,
        'use-db-taxa=s'            => \$use_db_taxa,
        'help'                     => \$help,
        'version'                  => \$ver);

    for ($summary_dir,$info_dir,$bz2_dir,$fasta_dir,$blast_db_dir,$blat_db_dir,$bwa_db_dir,$last_db_dir,
         $tax_db_dir,$cdata_dir,$cgraphs_dir,$cwords_dir,$krona_tools_dir) { $_ = fix_separator($_); }
}

sub print_help
{
    if ($ver or $help) { print (($ver ? $ver_str : ''), ($help ? $help_str : '')); exit; }
}

sub read_genome_info
{
    %genomes = ();
    my $file = "$summary_dir/list.hash";
    if (-e $file) { %genomes = %{retrieve($file)}; }
}

sub report_about_loaded_genomes
{
    my $total_size = sum map { $genomes{$_}->{'fna_size'} } keys %genomes;
    print "Info for ", commify(scalar keys %genomes), " genomes found, ", commify($total_size), " bytes in total\n";
}

sub load_genome_taxonomy
{
    %nodes              = %{retrieve(file_exists("$summary_dir/nodes-with-genome.hash"        ))};
    %name_to_taxid      = %{retrieve(file_exists("$summary_dir/genome-name-to-taxid.hash"     ))};
    %safe_name_to_taxid = %{retrieve(file_exists("$summary_dir/genome-safe-name-to-taxid.hash"))};
}

sub taxonomy_name_to_id
{
    my ($name) = @_;
    return exists($name_to_taxid{$name}) ? $name_to_taxid{$name} : exists($safe_name_to_taxid{$name}) ? $safe_name_to_taxid{$name} : 0;
}

sub find_most_recent_common_ancestor
{
    my ($t1,$t2) = @_;
    my %lc1 = ();
    my @la1 = ();
    for (my $t = $t1; $t != 1; $t = $nodes{$t}->{'p'}) { $lc1{$t} = 1; push @la1, $t; }
    my @la2 = ();
    for (my $t = $t2; $t != 1; $t = $nodes{$t}->{'p'}) { push @la2,$t; }
    my $i = scalar(@la2)-1;
    while ($i > 0 and exists $lc1{$la2[$i-1]}) { $i--; }
    my $r = (exists $lc1{$la2[$i]}) ? $la2[$i] : 1;
    return $r;
}

sub mark_genomes_to_use_by_taxa
{
    if (defined $use_db_taxa)
    {
        for (keys %genomes) { $genomes{$_}->{'use'} = 0; }
        foreach my $name (split(/;;/,$use_db_taxa))
        {
            my $use = 1;
            my $s = substr($name,0,1);
            if ($s eq '+') { $name = substr($name,1); }
            elsif ($s eq '-') { $name = substr($name,1); $use = 0; }
            my $id = taxonomy_name_to_id($name);
            if (!$id) { die "Unknown name: $name\n"; }
            set_genome_use_flag_in_subtree($id,$use);
        }
    }
    else { for (keys %genomes) { $genomes{$_}->{'use'} = 1; } }
}

sub set_genome_use_flag_in_subtree
{
    my ($id,$use) = @_;
    for (@{$nodes{$id}->{'g'}}) { $genomes{$_}->{'use'}  = $use; }
    for (@{$nodes{$id}->{'c'}}) { set_genome_use_flag_in_subtree($_,$use); }
}

sub tax_path_to_info_path
{
    return join('/',map({"0-$_"} split('/',$_[0])));
}

sub process_stdinout_or_dir
{
    my ($process_stream,$indir,$outdir,$n_files_in_parallel) = @_;
    if ((defined($indir) and $indir ne '') or (defined($outdir) and $outdir ne ''))
    {
        if (!defined($indir) or $indir eq '') { die "Input directory is not specified (--indir option)\n"; }
        if (!defined($outdir) or $outdir eq '') { die "Output directory is not specified (--outdir option)\n"; }
        if (!-e $indir) { die "Can't find input directory\n"; }
        if (!-d $indir) { die "Input directory is not a directory\n"; }
        if (!-e $outdir) { make_path($outdir); if (!-e $outdir) { die "Can't create output directory \"$outdir\"\n"; } }
        if (!-d $outdir) { die "Output directory is not a directory\n"; }

        my $indir_length = length($indir);
        my @files = ();
        find_files(\@files,\$outdir,$indir_length,$indir);
        my $n_files = scalar(@files);
        print STDERR "Processing $n_files files\n";

        my $pm = new Parallel::ForkManager($n_files_in_parallel);
        foreach my $infile (@files)
        {
            if (!-e $infile) { die "Can't find input file \"$infile\"\n"; }
            $pm->start and next;
            my $outfile = $outdir . substr($infile,$indir_length);
            open(my $IN,'<',$infile) or die "Can't open \"$infile\"\n";
            open(my $OUT,'>',$outfile) or die "Can't open \"$outfile\"\n";
            binmode $IN;
            binmode $OUT;
            &{$process_stream}($IN,$OUT);
            $pm->finish;
        }
        $pm->wait_all_children;
    }
    else
    {
        binmode STDIN;
        binmode STDOUT;
        &{$process_stream}(*STDIN{IO},*STDOUT{IO});
    }
}

sub find_files
{
    my ($arr,$outdir,$indir_length,$dir) = @_;
    my $dir_pattern = $dir;
    $dir_pattern =~ s/([\[\]])/\\$1/g;
    foreach my $entry (bsd_glob("$dir_pattern/*"))
    {
        my $out = $$outdir . substr($entry,$indir_length);
        if (-d $entry) { mkdir($out); find_files($arr,$outdir,$indir_length,$entry); }
        elsif (!-e $out) { push @{$arr}, $entry; }
    }
}

sub change_name_to_safe_name
{
    my ($name) = @_;
    $name =~ s/_/_5F/g;
    $name =~ s/"/_22/g;
    $name =~ s/%/_25/g;
    $name =~ s/&/_26/g;
    $name =~ s/'/_27/g;
    $name =~ s/\*/_2A/g;
    $name =~ s/\//_2F/g;
    $name =~ s/:/_3A/g;
    $name =~ s/</_3C/g;
    $name =~ s/>/_3E/g;
    $name =~ s/\?/_3F/g;
    $name =~ s/\[/_5B/g;
    $name =~ s/\\/_5C/g;
    $name =~ s/\]/_5D/g;
    $name =~ s/\|/_7C/g;
    return $name;
}

sub unescape_file_name
{
    my ($name) = @_;
    $name =~ s/_5F/_/g;
    $name =~ s/_22/"/g;
    $name =~ s/_25/%/g;
    $name =~ s/_26/&/g;
    $name =~ s/_27/'/g;
    $name =~ s/_2A/*/g;
    $name =~ s/_2F/\//g;
    $name =~ s/_3A/:/g;
    $name =~ s/_3C/</g;
    $name =~ s/_3E/>/g;
    $name =~ s/_3F/?/g;
    $name =~ s/_5B/[/g;
    $name =~ s/_5C/\\/g;
    $name =~ s/_5D/]/g;
    $name =~ s/_7C/|/g;
    return $name;
}

sub escape_spaces
{
    my ($text) = @_;
    $text =~ s/\s/_20/g;
    return $text;
}

sub fix_separator
{
    my ($text) = @_;
    $text =~ s/[\/\\]/$sep/g;
    return $text;
}

sub unix_separator
{
    my ($text) = @_;
    $text =~ s/\\/\//g;
    return $text;
}

sub file_exists
{
    my ($f) = @_;
    if (-e $f and -f $f) { return $f; }
    else { die "Error: Can't find file \"$f\""; }
}

sub dir_exists
{
    my ($n,$d) = @_;
    if (-e $d and -d $d) { return $d; }
    else { die "Error: Can't find $n directory \"$d\""; }
}

sub file_mtime
{
    my ($f) = @_;
    return (stat($f))[9];
}

sub commify
{
    my $text = reverse $_[0];
    $text =~ s/(\d\d\d)(?=\d)(?!\d*\.)/$1,/g;
    return scalar reverse $text;
}

1;
