#!/usr/bin/env perl
#
# This script is a part of Genome Search Toolkit
# Copyright (c) 2015 Kirill Kryukov
# see LICENSE.txt for details.
#
# Reads a directory with either blast output files (produced with '-outfmt 6').
# or last tabular format files.
#
# Produces an alignment score list.
#
# Input files must be ordered by query sequence, in the same order as in the query file.
#

use strict;
use warnings;
use File::Basename qw(basename);
use File::Find;
use Getopt::Long qw(GetOptionsFromArray :config pass_through);
use FindBin;
use lib "$FindBin::Bin";
use gst_utils;

gst_init();

my ($max_hits,$dir,$query,$queries_per_round,$query_list_file,$keep_accessions,$use_second_best);
my ($min_score) = (0);
GetOptionsFromArray(\@opts,
    'dir=s'               => \$dir,
    'query=s'             => \$query,
    'max-hits=i'          => \$max_hits,
    'min-score=f'         => \$min_score,
    'queries-per-round=i' => \$queries_per_round,
    'query-list=s'        => \$query_list_file,
    'accessions'          => \$keep_accessions,
    'use-second-best'     => \$use_second_best
);
if (@opts) { die "Unrecognized option \"$opts[0]\"\n"; }

$help_str .= qq{Options:
    --dir <DIR>             - Specify a directory with SAM files
    --query <FILE>          - Specify query file
    --minimal               - Input has only query id and score
    --max-hits <N>          - Output at most <N> hits per query (default: no limit)
    --min-score <S>         - Output only hits with at least this score (default: 0)
    --queries-per-round <N> - Process in rounds of N queries each
    --query-list <FILE>     - File with the list of queries to use (one name per line)
    --accessions            - Include genome accessions in the output
    --use-db-taxa <LIST>    - Genome taxa to include (;;-separated)
    --help                  - Show help
    --version               - Show version
};
if (!defined($dir) and !defined($query)) { $ver = 1; $help = 1; }
print_help();

if (!defined($dir)) { die "Search directory is not specified (--dir option)\n"; }
if (!defined($query)) { die "Query file is not specified (--query option)\n"; }

if (!defined($summary_dir)) { die "Genome summary directory is not specified (--genome-summary-dir option)\n"; }
if ($summary_dir eq '' or !-e $summary_dir or !-d $summary_dir) { die "Can't find genome summary directory \"$summary_dir\"\n"; }

if (!-e $dir) { die "Can't find directory \"$dir\"\n"; }
if (!-d $dir) { die "\"$dir\" is not a directory\n"; }
if (!-e $query) { die "Can't find query file \"$query\"\n"; }
if (!-f $query) { die "\"$query\" is not a file\n"; }

if (defined($queries_per_round) and $queries_per_round <= 0) { die "Queries per round is not a positive number\n"; }

read_genome_info();
if (defined($use_db_taxa)) { load_genome_taxonomy(); }
mark_genomes_to_use_by_taxa();

my @qnames = ();
my %qname_to_index = ();
my $nq = 0;
open (my $Q,'<',$query) or die "Can't open \"$query\"\n";
binmode $Q;
while (<$Q>)
{
    if (substr($_,0,1) eq '>')
    {
        s/[\x0D\x0A]+$//;
        my $qname = substr($_,1);
        $qnames[$nq] = $qname;
        $qname_to_index{$qname} = $nq;
        $nq++;
    }
}
close $Q;

my %use_query;
if ($query_list_file)
{
    open(my $QL,'<',$query_list_file) or die "Can't open \"$query_list_file\"\n";
    binmode $QL;
    while (<$QL>)
    {
        s/[\x0D\x0A]+$//;
        if (!exists $qname_to_index{$_}) { die "Unknown name in the query list file: \"$_\"\n"; }
        $use_query{$_} = 1;
    }
    close $QL;
    print STDERR "Using ", commify(scalar(keys %use_query)), " out of ", commify($nq), " queries\n";
}
else
{
    foreach my $qname (keys %qname_to_index) { $use_query{$qname} = 1; }
    print STDERR "Using all ", commify($nq), " queries\n";
}

if (!defined $queries_per_round) { $queries_per_round = $nq; }
binmode STDOUT;

my @q_best_score;
my @q_best_score_2;
my @q;
my @q_2;
my ($n_files,$n_hits) = (0,0);
my ($first_query_index,$last_query_index);
my %bad_files = ();
my %unknown_genomes = ();
my %genome_name_to_index;
my %genome_index_to_name;

for ($first_query_index = 0; $first_query_index < $nq; $first_query_index += $queries_per_round)
{
    $last_query_index = $first_query_index + $queries_per_round - 1;
    if ($last_query_index >= $nq) { $last_query_index = $nq - 1; }
    @q_best_score = ();
    @q_best_score_2 = ();
    @q = ();
    @q_2 = ();
    ($n_files) = (0);

    find({'wanted'=>\&{sub{ if (-f $_ and $_ =~ /\.(b6|lasttab)$/) { load_tabular_file($_); } }},'no_chdir'=>1},$dir);

    print STDERR ':';
    for (my $qi = $first_query_index; $qi <= $last_query_index; $qi++)
    {
        if (!exists $use_query{$qnames[$qi]}) { next; }

        print ' ', $qnames[$qi], "\n";
        my $n_hits_printed = 0;
        my $qq = $qi - $first_query_index;

        my $score = $q_best_score[$qq];
        foreach my $gindex (@{$q[$qq]})
        {
            print $score, ' ', $genome_index_to_name{$gindex}, "\n";
            $n_hits_printed++;
            if (defined($max_hits) and $n_hits_printed >= $max_hits) { last; }
        }

        if ($use_second_best)
        {
            my $score = $q_best_score_2[$qq];
            foreach my $gindex (@{$q_2[$qq]})
            {
                print $score, ' ', $genome_index_to_name{$gindex}, "\n";
                $n_hits_printed++;
                if (defined($max_hits) and $n_hits_printed >= $max_hits) { last; }
            }
        }
    }
}
if ($n_files) { print STDERR "\n"; }

sub load_tabular_file
{
    my ($file) = @_;

    my $gname = basename($file);
    if ($gname !~ /\.([^\.]+)$/) { die "Unknown format of file \"$file\"\n"; }
    my $format = $1;
    if ($format ne 'b6' and $format ne 'lasttab') { die "Unsupported format \"$format\"\n"; }

    $gname =~ s/\.[^\.]+$//;
    if (!exists $genomes{$gname}) { $unknown_genomes{$gname} = 1; return; }
    if (!$genomes{$gname}->{'use'}) { return; }

    my $genome_name = $gname;
    $genome_name =~ s/\s*\[[^\[\]]+\]$//;
    $genome_name =~ s/\s*\[[^\[\]]+\]$//;
    if ($keep_accessions)
    {
        $gname =~ /\[\S+\s(\S+)\s\d{4}-\d{2}-\d{2}\]$/ or die "Can't parse genome name \"$gname\"\n";
        my $accession = $1;
        $genome_name .= (' ' . unescape_file_name($accession));
    }

    my $genome_index;
    if (exists $genome_name_to_index{$genome_name}) { $genome_index = $genome_name_to_index{$genome_name}; }
    else
    {
        $genome_index = scalar(keys %genome_name_to_index) + 1;
        $genome_name_to_index{$genome_name} = $genome_index;
        $genome_index_to_name{$genome_index} = $genome_name;
    }

    open(my $F,'<',$file) or die "Can't open \"$file\"\n";
    binmode $F;
    if (exists $genomes{$gname}->{'pos'}) { seek($F,$genomes{$gname}->{'pos'},0); }
    my $last_pos = tell($F);
    while (<$F>)
    {
        s/[\x0D\x0A]+$//;
        my @fields = split(/\s+/,$_);

        if (scalar(@fields) != 2) { $bad_files{$file} = 1; next; }
        my ($qname,$as) = @fields[0,1];

        if (!exists $qname_to_index{$qname}) { warn "Unknown query \"$qname\" in \"$file\"\n"; next; }
        if (!exists $use_query{$qname}) { next; }

        my $qi = $qname_to_index{$qname};
        if ($qi < $first_query_index) { die "qi: $qi < first_query_index: $first_query_index (gname: $gname)\n"; }
        if ($qi > $last_query_index) { last; }
        $last_pos = tell($F);

        if ($as !~ /^(\d+|\d+\.\d+)$/) { $bad_files{$file} = 1; next; }
        if ($as < $min_score) { next; }

        my $qq = $qi - $first_query_index;

        if (!defined $q_best_score[$qq])
        {
            $q_best_score[$qq] = $as;
            @{$q[$qq]} = ($genome_index);
            $n_hits++;
            next;
        }

        if ($as > $q_best_score[$qq])
        {
            if ($use_second_best)
            {
                $q_best_score_2[$qq] = $q_best_score[$qq];
                @{$q_2[$qq]} = @{$q[$qq]};
            }
            $q_best_score[$qq] = $as;
            @{$q[$qq]} = ($genome_index);
            $n_hits++;
            next;
        }

        if ($as == $q_best_score[$qq])
        {
            push @{$q[$qq]}, $genome_index;
            $n_hits++;
            next;
        }

        if (!$use_second_best) { next; }

        if (!defined $q_best_score_2[$qq])
        {
            $q_best_score_2[$qq] = $as;
            @{$q_2[$qq]} = ($genome_index);
            $n_hits++;
            next;
        }

        if ($as > $q_best_score_2[$qq])
        {
            $q_best_score_2[$qq] = $as;
            @{$q_2[$qq]} = ($genome_index);
            $n_hits++;
            next;
        }

        if ($as == $q_best_score_2[$qq])
        {
            push @{$q_2[$qq]}, $genome_index;
            $n_hits++;
            next;
        }
    }
    $genomes{$gname}->{'pos'} = $last_pos;
    close $F;
    $n_files++;
}

print STDERR ' - ', commify($n_files), ' files processed, loaded ', commify($n_hits), " hits\n";
my $n_bad_files = scalar keys %bad_files;
if ($n_bad_files) { print STDERR " - $n_bad_files corrupted files:\n    ", join("\n    ",keys %bad_files), "\n"; }
my $n_unknown_genomes = scalar keys %unknown_genomes;
if ($n_unknown_genomes) { print STDERR " - $n_unknown_genomes unknown genomes:\n    ", join("\n    ",sort keys %unknown_genomes), "\n"; }

my $elapsed_time = time - $start_time;
print STDERR " - done, $elapsed_time second" . (($elapsed_time==1)?'':'s') . " elapsed\n";
