#!/usr/bin/env perl
#
# This script is a part of Genome Search Toolkit
# Copyright (c) 2016 Kirill Kryukov
# see LICENSE.txt for details.
#
# Divides a list of genomes into parts of approximately equal size (greedily).
# The input genome names come from standard input, one per line.
# The output lists are stored in N files (N is the number of parts).
#

use strict;
use warnings;
use File::Basename qw(basename fileparse);
use File::Find;
use File::Path qw(make_path);
use Getopt::Long qw(GetOptionsFromArray :config pass_through);
use FindBin;
use lib "$FindBin::Bin";
use gst_utils;

gst_init();

my ($strict,$n_parts,$prefix,$suffix) = (0,0,'','');
GetOptionsFromArray(\@opts,
    'n-parts=i' => \$n_parts,
    'prefix=s'  => \$prefix,
    'strict'    => \$strict
);
if (@opts) { die "Unrecognized option \"$opts[0]\"\n"; }

$help_str .= qq{Options:
    --n-parts <N> - Partition into <N> parts
    --prefix <S>  - Use <S> as output file name prefix
    --suffix <E>  - Add <E> as output file name suffix (eg: '.txt')
    --strict      - Strict mode - stop on any unrecognized genome
    --help        - Show help
    --version     - Show version
};
if (!defined $n_parts or $n_parts <= 0) { $ver = 1; $help = 1; }
print_help();

if ($n_parts < 1) { print STDERR "Number of parts should be at least 1\n"; exit; }

if ($prefix ne '') { my ($filename,$dirs,$suffix) = fileparse($prefix); make_path($dirs); }

read_genome_info();
print STDERR scalar(keys %genomes), " genomes are installed\n";
if (defined($use_db_taxa)) { load_genome_taxonomy(); }

my ($n_used_genomes,$n_unknown_genomes) = (0,0);
binmode STDIN;
while (<STDIN>)
{
    s/[\x0D\x0A]+$//;
    my $name = $_;
    if (exists $genomes{$name}) { $genomes{$name}->{'use'} = 1; $n_used_genomes++; }
    else
    {
        if ($name =~ /^(.+?)\s+\[([^\[\]]+)\]$/)
        {
            my $safe_name = change_name_to_safe_name($1) . ' [' . change_name_to_safe_name($2) . ']';
            if (exists $genomes{$safe_name}) { $genomes{$safe_name}->{'use'} = 1; $n_used_genomes++; }
            else { if ($strict) { die "Unknown genome \"$name\"\n"; } $n_unknown_genomes++; }
        }
        else { if ($strict) { die "Can't parse genome name \"$name\"\n"; } $n_unknown_genomes++; }
    }
}
if ($n_unknown_genomes) { print STDERR "Error: Skipping $n_unknown_genomes unknown genomes\n"; }
print STDERR "Dividing $n_used_genomes genomes into $n_parts parts\n";

my @parts = ();
for (my $i=0; $i<$n_parts; $i++) { $parts[$i]->{'size'} = 0; }
foreach my $gname (sort { $genomes{$b}->{'fna_size'} <=> $genomes{$a}->{'fna_size'} } grep { $genomes{$_}->{'use'} } keys %genomes)
{
    my $i = find_smallest_part();
    push @{$parts[$i]->{'genomes'}}, $gname;
    $parts[$i]->{'size'} += $genomes{$gname}->{'fna_size'};
}

for (my $i=0; $i<$n_parts; $i++)
{
    my $file = $prefix . sprintf('%d',$i+1);
    open(my $F,'>',$file) or die "Can't create \"$file\"\n";
    binmode $F;
    foreach my $gname (@{$parts[$i]->{'genomes'}}) { print $F $gname, "\n"; }
    close $F;
}

sub find_smallest_part
{
    my ($si,$ss) = (-1,-1);
    for (my $i=0; $i<$n_parts; $i++)
    {
        if ($ss < 0 or $parts[$i]->{'size'} < $ss) { ($si,$ss) = ($i,$parts[$i]->{'size'}); }
    }
    return $si;
}
