#!/usr/bin/env perl
#
# This script is a part of Genome Search Toolkit
# Copyright (c) 2016 Kirill Kryukov
# see LICENSE.txt for details.
#
# Reads reduced search output.
# For each query sequence, computes taxon specificity, for all taxa it is specific to.
# (Taxon specificity is defined as the difference between the best scores inside and outside the taxon)
#
# Then it prints number of specific query sequences for each taxon, for a given specificity range.
#

use strict;
use warnings;
use Getopt::Long;
use List::Util qw(min max);
use POSIX qw(floor ceil);
use Storable;

my ($start_time,$script_name,$script_version,$script_date,$script_years) = (time,'taxon-specificity-all','0.1.0','2016-03-04','2016');
$| = 1;

my ($min_specificity,$max_specificity,$specificity_step,$list_of_specificities_str,$infile,$query,$summary_dir,$random_score,$dot_n,$max_queries_to_process,$quiet,$help,$ver);
GetOptions(
    'min-spec=f'           => \$min_specificity,
    'max-spec=f'           => \$max_specificity,
    'spec-step=f'          => \$specificity_step,
    'spec-list=s'          => \$list_of_specificities_str,
    'in=s'                 => \$infile,
    'query=s'              => \$query,
    'genome-summary-dir=s' => \$summary_dir,
    'random-score=f'       => \$random_score,
    'quiet'                => \$quiet,
    'dot=i'                => \$dot_n,
    'max-queries=i'        => \$max_queries_to_process,
    'help'                 => \$help,
    'version'              => \$ver)
or die "Can't parse command line arguments\n";

my $ver_str = "$script_name, version $script_version, $script_date\nCopyright (c) $script_years Kirill Kryukov\n";
my $help_str = qq{Usage: $script_name <option>...
Options:
    --min-spec S               - Minimum taxon specificity
    --max-spec S               - Maximum taxon specificity
    --spec-step S              - Step of increasing taxon specificity
    --spec-list S1[,S2]        - List of taxon specificities
    --in <FILE>                - File with hit scores
    --query <FILE>             - Query file
    --genome-summary-dir <DIR> - Specify genome summary directory
    --random-score S           - Assume that score S means random match
    --max-queries N            - Process only at most N queries
    --dot N                    - Print a dot every N queries
    --quiet                    - No messages
    --help                     - Show help
    --version                  - Show version
};

print (($ver ? $ver_str : ''), ($help ? $help_str : ''));
if (!defined($min_specificity) and !defined($max_specificity) and !defined($specificity_step and !defined($list_of_specificities_str)) and
    !defined($infile) and !defined($query)) { if (!$help and !$ver) { print $ver_str, $help_str; } exit; }

if ( !defined($list_of_specificities_str) and
     !(defined($min_specificity) and defined($max_specificity) and defined($specificity_step)) )
    { die "Taxon specificities are not specified.\n" .
          "Either --spec-list must be provided, or --min-spec, --max-spec and --spec-step\n"; }
if (!defined($random_score)) { $random_score = 0; }
if (!defined($infile)) { die "Input file is not specified (--in option)\n"; }
if (!defined($query)) { die "Query file is not specified (--query option)\n"; }
if (!defined($summary_dir)) { die "Genome summary directory is not specified (--genome-summary-dir option)\n"; }

# Constructing sorted list of taxon specificities.
my @spec_list;
if (defined($list_of_specificities_str))
{
    @spec_list = sort { $a <=> $b } grep { /^\d+$/ } split(',',$list_of_specificities_str);
    if (!scalar(@spec_list)) { die "Can't find valid specificities in \"$list_of_specificities_str\"\n"; }
    $min_specificity = $spec_list[0];
    $max_specificity = $spec_list[scalar(@spec_list)-1];
}
else
{
    for (my $s = $min_specificity; $s <= $max_specificity; $s += $specificity_step) { push @spec_list, $s; }
}

my $n_specificities = scalar(@spec_list);

$infile = file_exists($infile);
$query = file_exists($query);
my %nodes              = %{retrieve(file_exists("$summary_dir/nodes-with-genome.hash"        ))};
my %name_to_taxid      = %{retrieve(file_exists("$summary_dir/genome-name-to-taxid.hash"     ))};
my %safe_name_to_taxid = %{retrieve(file_exists("$summary_dir/genome-safe-name-to-taxid.hash"))};

# Separate hash for storing parent node ids.
my %np;
foreach my $t (keys %nodes) { $np{$t} = $nodes{$t}->{'p'}; }

# Constructing hash for looking up taxonomic ID by name.
my %taxid_to_name;
foreach my $n (keys %name_to_taxid) { $taxid_to_name{$name_to_taxid{$n}} = $n; }
foreach my $n (keys %safe_name_to_taxid) { $taxid_to_name{$safe_name_to_taxid{$n}} = $n; }
my %all_name_to_taxid;
foreach my $t (keys %taxid_to_name) { $all_name_to_taxid{$taxid_to_name{$t}} = $t; }

# Indexed hash of names for faster lookup.
my @indexed_name_to_taxid;
foreach my $n (keys %all_name_to_taxid) { $indexed_name_to_taxid[ord(substr($n,0,1))]->{substr($n,1)} = $all_name_to_taxid{$n}; }

#my %lineage;
#foreach my $c (keys %np) { for (my $p=$c; defined($p); $p = $np{$p}) { push @{$lineage{$c}}, $p; } }

my (@qnames,%qname_to_index,@qlengths);
my ($nq,$qname) = (0,'');
open (my $Q,'<',$query) or die "Can't open query file \"$query\"\n";
binmode $Q;
while (<$Q>)
{
    s/[\x0D\x0A]+$//;
    if (substr($_,0,1) eq '>') { $qname = substr($_,1); $qnames[$nq] = $qname; $qname_to_index{$qname} = $nq; $nq++; next; }
    $qlengths[$nq-1] += length($_);
}
close $Q;
if (!$quiet) { print STDERR "$nq queries\n"; }

my %nq_by_taxon_by_spec = ();

my ($qi,$sname,%hit_scores) = (-1,'');
my $dot_next = $dot_n;
open (my $IN,'<',$infile) or die "Can't open \"$infile\"\n";
binmode $IN;
while (<$IN>)
{
    if (defined($max_queries_to_process) and $qi >= $max_queries_to_process) { last; }

    s/[\x0D\x0A]+$//;
    if (substr($_,0,1) eq ' ')
    {
        process_one_query();
        $qname = substr($_,1);
        $qi++;

        if (defined($dot_n) and $qi >= $dot_next) { print STDERR '.'; $dot_next += $dot_n; }

        if ($qname ne $qnames[$qi]) { die "Query name mismatch\n"; }
        %hit_scores = ();
        next;
    }

    if ($qi < 0) { next; }

    my $si = index($_,' ',0);
    my $score = substr($_,0,$si);
    my $ni = $si+1;
    my $sname = substr($_,$ni,rindex($_,' ')-$ni);
    my $sid = $indexed_name_to_taxid[ord(substr($sname,0,1))]->{substr($sname,1)};

    if (!$sid) { die "Unknown database name \"$sname\"\n"; }
    $hit_scores{$sid} = $score;
}
close $IN;
process_one_query();

sub process_one_query
{
    if ($qi < 0 or !scalar(keys %hit_scores)) { return; }

    #my $max_score = 0;
    #foreach my $sc (values %hit_scores) { if ($sc > $max_score) { $max_score = $sc; } }
    #foreach my $c (keys %hit_scores) { if ($hit_scores{$c} != $max_score) { delete $hit_scores{$c}; } }

    my %contains1;
    foreach my $c (keys %hit_scores)
    {
        for (my $p=$c; defined($p); $p = $np{$p}) { $contains1{$p}->{$c} = 1; }
        #foreach my $p (@{$lineage{$c}}) { $contains1{$p}->{$c} = 1; }
    }

    my $max_score = 0;
    while (my ($ht,$sc) = each %hit_scores) { if ($sc > $max_score) { $max_score = $sc; } }

    foreach my $t (keys %contains1)
    {
        my $best_in = $random_score;
        foreach my $ht (keys %{$contains1{$t}}) { if ($hit_scores{$ht} > $best_in) { $best_in = $hit_scores{$ht}; } }

        my $best_out;
        if ($best_in < $max_score) { $best_out = $max_score; }
        else
        {
            $best_out = $random_score;
            my $ch = \$contains1{$t};
            foreach my $ht (keys %hit_scores) { if ($hit_scores{$ht} > $best_out and !exists $$ch->{$ht}) { $best_out = $hit_scores{$ht}; } }
            #while (my ($ht,$sc) = each %hit_scores) { if ($sc > $best_out and !exists $$ch->{$ht}) { $best_out = $sc; } }
        }

        my $spec = sprintf('%.1f',$best_in - $best_out);
        my $ss = min($max_specificity,$spec);
        for (my $si = 0; $si < $n_specificities and $spec_list[$si] <= $ss; $si++) { $nq_by_taxon_by_spec{$t}->{$spec_list[$si]}++; }
    }
}

if (defined($dot_n)) { print STDERR "\n"; }
if (!defined($max_queries_to_process) and $qi+1 != $nq) { die "Query file has $nq sequences, but scores file has $qi\n"; }

# %taxon_total_n_reads is only used for sorting the taxa,
# in order to print the most abundantly present taxa at the top of the output file.
my %taxon_total_n_reads = ();
foreach my $t (keys %nq_by_taxon_by_spec)
{
    my $n = 0;
    foreach my $s (@spec_list) { if (exists $nq_by_taxon_by_spec{$t}->{$s}) { $n += $nq_by_taxon_by_spec{$t}->{$s}; } }
    $taxon_total_n_reads{$t} = $n;
}

print '# ';
foreach my $s (@spec_list) { print "$s\t"; }
print "TaxID\tTaxName\n";

my $total_n_queries = $qi + 1;
foreach my $s (@spec_list)
{
    my $n_specific = exists($nq_by_taxon_by_spec{1}->{$s}) ? $nq_by_taxon_by_spec{1}->{$s} : 0;
    my $n_unspecific = $total_n_queries - $n_specific;
    print "$n_unspecific\t";
}
print "0\t-\n";

foreach my $t (sort { $taxon_total_n_reads{$b} <=> $taxon_total_n_reads{$a} || $nodes{$a}->{'n'} cmp $nodes{$b}->{'n'} } keys %nq_by_taxon_by_spec)
{
    foreach my $s (@spec_list) { print ((exists($nq_by_taxon_by_spec{$t}->{$s}) ? $nq_by_taxon_by_spec{$t}->{$s} : '0'), "\t"); }
    print $t, "\t", $nodes{$t}->{'n'}, "\n";
}

if (!$quiet)
{
    my $elapsed_time = time - $start_time;
    print STDERR "All done, $elapsed_time second" . (($elapsed_time==1)?'':'s') . " elapsed\n";
}

sub file_exists { my ($f) = @_; if (-e $f and -f $f) { return $f; } else { die "Error: Can't find file \"$f\""; } }
