#!/usr/bin/env perl
#
# This script is a part of Genome Search Toolkit
# Copyright (c) 2015 Kirill Kryukov
# see LICENSE.txt for details.
#
# Reads reduced search output.
# For each query sequence and for each specified taxon, computes taxon-specificity,
# defined as the difference between the best scores inside and outside the taxon.
# Then it prints number of query sequences for each specificity value (allowing to plot a histogram).
#
# With '--rel' it uses specificities divided by the length of aligned part of query sequence.
# It then maps these ratios to range [-max ratio .. max ratio].
#

use strict;
use warnings;
use Getopt::Long;
use List::Util qw(max);
use POSIX qw(floor ceil);
use Storable;

my ($start_time,$script_name,$script_version,$script_date,$script_years) = (time,'taxon-specificity','0.1.0','2015-12-17','2015');
$| = 1;

my ($method,$resolution,$infile,$query,$taxon_list,$summary_dir,$non_negative,$positive,$random_score,$quiet,$help,$ver);
GetOptions(
    'abs'                  => sub { $method = 0; },
    'rel'                  => sub { $method = 1; },
    'res=s'                => \$resolution,
    'in=s'                 => \$infile,
    'query=s'              => \$query,
    'taxon=s'              => \$taxon_list,
    'genome-summary-dir=s' => \$summary_dir,
    'non-negative'         => \$non_negative,
    'positive'             => \$positive,
    'random-score=f'       => \$random_score,
    'quiet'                => \$quiet,
    'help'                 => \$help,
    'version'              => \$ver)
or die "Can't parse command line arguments\n";

my $ver_str = "$script_name, version $script_version, $script_date\nCopyright (c) $script_years Kirill Kryukov\n";
my $help_str = qq{Usage: $script_name --(abs|rel) <option>...
    --abs                      - Use absolute scores
    --rel                      - Use relative scores (score / read length)
Options:
    --res <N>                  - rel: Set resolution to N points between 0 and perfect score.
                                 abs: Use bin size of N.
    --in <FILE>                - File with hit scores
    --query <FILE>             - Query file
    --taxon <NAME>             - Taxons (;;-separated)
    --genome-summary-dir <DIR> - Specify genome summary directory
    --only-non-negative        - Output only non-negative specifities
    --only-positive            - Output only positive specifities
    --random-score S           - Assume that score S means random match
    --quiet                    - No messages
    --help                     - Show help
    --version                  - Show version
};

print (($ver ? $ver_str : ''), ($help ? $help_str : ''));
if (!defined($method) and !defined($resolution) and !defined($infile) and !defined($query) and !defined($taxon_list)) { if (!$help and !$ver) { print $ver_str, $help_str; } exit; }
if (!defined($method)) { die "Scoring method is not specified (--abs or --rel option)\n"; }
if (!defined($random_score)) { $random_score = 0; }
if (!defined($resolution) and $method == 1) { die "Resolution is not specified (--res option)\n"; }
if (!defined($resolution)) { $resolution = 100; }
if (!defined($infile)) { die "Input file is not specified (--in option)\n"; }
if (!defined($query)) { die "Query file is not specified (--query option)\n"; }
if (!defined($taxon_list)) { die "Taxon is not specified (--taxon option)\n"; }
if (!defined($summary_dir)) { die "Genome summary directory is not specified (--genome-summary-dir option)\n"; }

$infile = file_exists($infile);
$query = file_exists($query);
my %nodes              = %{retrieve(file_exists("$summary_dir/nodes-with-genome.hash"        ))};
my %name_to_taxid      = %{retrieve(file_exists("$summary_dir/genome-name-to-taxid.hash"     ))};
my %safe_name_to_taxid = %{retrieve(file_exists("$summary_dir/genome-safe-name-to-taxid.hash"))};

my @taxons = split(';;',$taxon_list);
my @taxids = map { taxonomy_name_to_id($_) or die "Unknown taxonomic name \"$_\"\n"; } @taxons;
my $n_taxons = scalar @taxons;
if (!$n_taxons) { die "No taxa were specified\n"; }

my (@qnames,%qname_to_index,@qlengths);
my ($nq,$qname) = (0,'');
open (my $Q,'<',$query) or die "Can't open query file \"$query\"\n";
binmode $Q;
while (<$Q>)
{
    s/[\x0D\x0A]+$//;
    if (substr($_,0,1) eq '>') { $qname = substr($_,1); $qnames[$nq] = $qname; $qname_to_index{$qname} = $nq; $nq++; next; }
    $qlengths[$nq-1] += length($_);
}
close $Q;
if (!$quiet) { print STDERR "$nq queries\n"; }

#my $debug_read_name = '21448';
#my $debug_read_index = -1;;

my (@best_in,@best_out);
my ($qi,$sname,$max_score_per_bp,$max_score_per_bp__score,$max_score_per_bp__len) = (-1,'',0,0,0);
open (my $IN,'<',$infile) or die "Can't open \"$infile\"\n";
binmode $IN;
while (<$IN>)
{
    s/[\x0D\x0A]+$//;
    if (substr($_,0,1) eq ' ')
    {
        $qname = substr($_,1);
        $qi++;
        if ($qname ne $qnames[$qi]) { die "Query name mismatch\n"; }
        #if ($qname eq $debug_read_name) { $debug_read_index = $qi; }

        if ($method == 1)
        {
            for (my $ti=0; $ti<$n_taxons; $ti++)
            {
                $best_in [$ti][$qi][0] = $random_score;
                $best_out[$ti][$qi][0] = $random_score;
                $best_in [$ti][$qi][1] = 1;
                $best_out[$ti][$qi][1] = 1;
            }
        }
        else
        {
            for (my $ti=0; $ti<$n_taxons; $ti++)
            {
                $best_in [$ti][$qi] = $random_score;
                $best_out[$ti][$qi] = $random_score;
            }
        }

        next;
    }

    if ($qi < 0) { next; }
    /^(\d+|\d+\.\d+)\s+\S+\s+\S+\s+(\d+)\s+(\S.+?)\s+\S+$/ or die "Can't parse \"$_\"\n";
    my ($score,$qalen,$sname,$sid) = ($1,$2,$3,taxonomy_name_to_id($3));
    if (!$sid) { die "Unknown database name \"$sname\"\n"; }

    my @within = (0) x $n_taxons;
    for (my $id = $sid; defined $id; $id = $nodes{$id}->{'p'})
    {
        for (my $ti=0; $ti<$n_taxons; $ti++) { if ($id eq $taxids[$ti]) { $within[$ti] = 1; } }
    }
    for (my $ti=0; $ti<$n_taxons; $ti++)
    {
        if ($method == 1)
        {
            if ($within[$ti]) { if ($score > $best_in [$ti][$qi][0]) { $best_in [$ti][$qi][0] = $score; $best_in [$ti][$qi][1] = $qalen; } }
                         else { if ($score > $best_out[$ti][$qi][0]) { $best_out[$ti][$qi][0] = $score; $best_out[$ti][$qi][1] = $qalen; } }
        }
        else
        {
            if ($within[$ti]) { if ($score > $best_in [$ti][$qi]) { $best_in [$ti][$qi] = $score; } }
                         else { if ($score > $best_out[$ti][$qi]) { $best_out[$ti][$qi] = $score; } }
        }
    }
    my $score_per_bp = $score / $qalen;
    if ($score_per_bp > $max_score_per_bp) { $max_score_per_bp = $score_per_bp; $max_score_per_bp__score = $score; $max_score_per_bp__len = $qalen; }
}
close $IN;
if ($qi+1 != $nq) { die "Query file has $nq sequences, but scores file has $qi\n"; }
if (!$quiet and $method == 1) { print STDERR 'Max score/bp ratio: ', sprintf('%.2f',$max_score_per_bp), " score units / bp\n"; }

my @hist;
my ($n_no_hits,$n_have_hits,$min_d,$max_d) = (0,0,1000000,-1000000);
for (my $i=0; $i<$nq; $i++)
{
    #if ($i == $debug_read_index)
    #{
    #    print STDERR "Read $i, name: $debug_read_name\n";
    #    if ($method == 1) {}
    #    else
    #    {
    #        print STDERR "    Best in score:  $best_in[0][$i]\n";
    #        print STDERR "    Best out score: $best_out[0][$i]\n";
    #    }
    #}

    if ($method == 1) { if ($best_in[0][$i][0] <= $random_score and $best_out[0][$i][0] <= $random_score) { $n_no_hits++; next; } }
    else { if ($best_in[0][$i] <= $random_score and $best_out[0][$i] <= $random_score) { $n_no_hits++; next; } }

    $n_have_hits++;
    for (my $ti=0; $ti<$n_taxons; $ti++)
    {
        my $d_int;

        if ($method == 1)
        {
            my $d = $best_in[$ti][$i][0] - $best_out[$ti][$i][0];
            my $qalen = max($best_in[$ti][$i][1],$best_out[$ti][$i][1]);
            $d_int = sprintf('%.0f',$d / ($qalen * $max_score_per_bp) * $resolution);
        }
        else
        {
            my $d = $best_in[$ti][$i] - $best_out[$ti][$i];
            #$d_int = sprintf('%.0f',$d / $resolution);
            $d_int = $d / $resolution;
            if ($d_int < 0) { $d_int = ceil($d_int); }
            else { $d_int = floor($d_int); }
        }

        if ($d_int eq '-0') { $d_int = 0; }
        $hist[$ti]->{$d_int}++;
        if ($d_int < $min_d) { $min_d = $d_int; }
        if ($d_int > $max_d) { $max_d = $d_int; }
    }
}

if ($min_d > 0) { $min_d = 0; }
if ($max_d < 0) { $max_d = 0; }

for (my $d=$min_d; $d<=$max_d; $d++)
{
    for (my $ti=0; $ti<$n_taxons; $ti++) { if (!exists $hist[$ti]->{$d}) { $hist[$ti]->{$d} = 0; } }
}

my @hist_at_least;
for (my $ti=0; $ti<$n_taxons; $ti++)
{
    $hist_at_least[$ti]->{$max_d} = $hist[$ti]->{$max_d};
    for (my $d=$max_d-1; $d>=$min_d; $d--) { $hist_at_least[$ti]->{$d} = $hist_at_least[$ti]->{$d+1} + $hist[$ti]->{$d}; }
}

for (my $d = $positive ? 1 : $non_negative ? 0 : $min_d; $d<=$max_d; $d++)
{
    if ($method == 1) { printf("%9.2f",$d*100/$resolution); }
    else { printf("%.0f",$d*$resolution); }
    for (my $ti=0; $ti<$n_taxons; $ti++) { print sprintf("\t%d",$hist_at_least[$ti]->{$d}); }
    print "\n";
}
if (!$quiet) { print STDERR "$n_have_hits queries have hits, $n_no_hits queries have no hits\n"; }
#if ($method == 0) { print STDERR "Approximate score range: ", $min_d*$resolution, ' .. ', $max_d*$resolution, "\n"; }


sub file_exists { my ($f) = @_; if (-e $f and -f $f) { return $f; } else { die "Error: Can't find file \"$f\""; } }

sub taxonomy_name_to_id
{
    my ($name) = @_;
    return exists($name_to_taxid{$name}) ? $name_to_taxid{$name} : exists($safe_name_to_taxid{$name}) ? $safe_name_to_taxid{$name} : 0;
}
